/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the G100 and G200 video extension
 * registers.  
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/i2c.h>
#include "mgavideo.h"

#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */

#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif


/* i2c registers */
#define PALWTADD        0x3c00
#define X_DATAREG       0x3c0a
#define XMISCCTRL       0x1e
#define XGENIOCTRL      0x2a
#define XGENIODATA      0x2b
#define SDA             0x10
#define SCL             0x20

#define I2C_DELAY       10


/* video registers */
#define VINCTL0         0x3e00
#define VINCTL1         0x3e04
#define VBIADDR0        0x3e08
#define VBIADDR1        0x3e0c
#define VINADDR0        0x3e10
#define VINADDR1        0x3e14
#define VINNEXTWIN      0x3e18
#define VINCTL          0x3e1c
#define VSTATUS         0x3e30
#define VICLEAR         0x3e34
#define VIEN            0x3e38

/* CODEC registers */
#define CODECCTL        0x3e40
#define CODECADDR       0x3e44
#define CODECHOSTPTR    0x3e48
#define CODECHARDPTR    0x3e4c
#define CODECLCODE      0x3e50

/* Backend scaler registers */
#define BESA1ORG        0x3d00
#define BESA2ORG        0x3d04
#define BESB1ORG        0x3d08
#define BESB2ORG        0x3d0c
#define BESA1CORG       0x3d10
#define BESA2CORG       0x3d14
#define BESB1CORG       0x3d18
#define BESB2CORG       0x3d1c
#define BESCTL          0x3d20
#define BESPITCH        0x3d24
#define BESHCOORD       0x3d28
#define BESVCOORD       0x3d2c
#define BESHISCAL       0x3d30
#define BESVISCAL       0x3d34
#define BESHSRCST       0x3d38
#define BESHSRCEND      0x3d3c
#define BESV1WGHT       0x3d48
#define BESV2WGHT       0x3d4c
#define BESHSRCLST      0x3d50
#define BESV1SRCLST     0x3d54
#define BESV2SRCLST     0x3d58
#define BESGLOBCTL      0x3dc0
#define BESSTATUS       0x3dc4


#define VINCTLx_MASK    0x00000fff
#define VBIADDRx_MASK   0x00fff200
#define VINADDRx_MASK   0x00fffff8


/* mga video status */
#define VI_IVSYNC    0x0001 /* input vsync interrupt */
#define VI_CMDCMPL   0x0002 /* command complete interrupt */
#define VI_BLVL      0x0004 /* buffer level interrupt */
#define VI_EOI       0x0008 /* codec end of image interrupt */
#define VS_FIELD     0x0100 /* field 0 or 1 detected */
#define VS_VIDEO     0x0200 /* indicates video was captured */
#define VS_RAWVBI    0x0400 /* indicates raw vbi was captured */
#define VS_SLICEDVBI 0x0800 /* indicates sliced vbi was captured */
#define VIN_OFF      0x0
#define VIN_ON       0x1
#define VBI_OFF      0x0
#define VBI_RAW      0x2
#define VBI_SLICED   0x6


/* video attributes - these should probably be configurable on the fly
 *                    so users with different desktop sizes can keep
 *                    captured data off the desktop
 */
#define _VINPITCH 704
#define _VINADDR0 0x200000
#define _VBIADDR0 0x2c0000
#define _VINADDR1 0x300000
#define _VBIADDR1 0x3c0000
#define _VINWID   704
#define _VINHGT   290


/****************************************************************************
* mga_dev : represents one mga board.  since we can be dealing with multiple
*           boards, all entry points to this module must specify a board
****************************************************************************/
struct mga_dev
{
        struct i2c_bus i2c;

        u8* fb;
        u8* ctrl;

        int inuse;
        char name[16];

        /* device info from pci */
        u32 device;
        int irq;  

        /* caps */
        int hasaudio;
        int hastuner;
        int hasdecoder;

        /* status of the video grabber */
        u32 status;
        u32 locked;

        /* registered bottom half of isr */
        struct tq_struct tqnode_dpc;
        int req_even;
        int req_odd;
        int irq_status;

        /* statistics */
        int grabbed;
};


/****************************************************************************
* raw register access : these routines directly interact with the mga's
*                       control aperature.  must not be called until after
*                       the board's pci memory has been mapped.
****************************************************************************/
static u32 _mgaread( struct mga_dev* mga, u32 reg )
{
        return readl( mga->ctrl + reg );
}

static void _mgawrite( struct mga_dev* mga, u32 reg, u32 data )
{
        writel( data, mga->ctrl + reg );
}


/****************************************************************************
* mga i2c bus interface
****************************************************************************/
static void _mgai2c_attach( struct i2c_bus* bus, int id )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        dprintk( "mgavideo: i2c device attach: %d\n", id );

        switch( id )
        {
        case I2C_DRIVERID_MSP3400:
                mga->hasaudio = 1;
                break;

        case I2C_DRIVERID_TUNER:
                mga->hastuner = 1;
                break;

        case I2C_DRIVERID_KS0127:
                mga->hasdecoder = 1;
                break;
        }
}

static void _mgai2c_detach( struct i2c_bus* bus, int id )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        dprintk( "mgavideo: i2c device detach: %d\n", id );

        switch( id )
        {
        case I2C_DRIVERID_MSP3400:
                mga->hasaudio = 0;
                break;

        case I2C_DRIVERID_TUNER:
                mga->hastuner = 0;
                break;

        case I2C_DRIVERID_KS0127:
                mga->hasdecoder = 0;
                break;
        }
}

static void _mgai2c_setlines( struct i2c_bus* bus, int ctrl, int data )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        if( ctrl )
        {
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SCL, mga->ctrl + X_DATAREG );
        }
        else
        {
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SCL, mga->ctrl + X_DATAREG );
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) | SCL, mga->ctrl + X_DATAREG );
        }

        if( data )
        {
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        }
        else
        {
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) | SDA, mga->ctrl + X_DATAREG );
        }

        udelay( I2C_DELAY );
}

static int _mgai2c_getdataline( struct i2c_bus* bus )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        return ( readb( mga->ctrl + X_DATAREG ) & SDA )?1:0;
}


/****************************************************************************
* private api : internal interface to the mga's video subsystem
****************************************************************************/


/****************************************************************************
* mgavideo api : public interface to the mga's video subsystem
****************************************************************************/
/* component device control */
void mgavideo_tuner( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga == NULL )
                return;

        if( mga->hastuner )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_TUNER, cmd, arg );
}

void mgavideo_audio( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga == NULL )
                return;

        if( mga->hasaudio )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_MSP3400, cmd, arg );
}

void mgavideo_decoder( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga == NULL )
                return;

        if( mga->hasdecoder )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_KS0127, cmd, arg );
}

int mgavideo_querycaps( struct mga_dev* mga )
{
        int caps = 0;

        if( mga == NULL )
                return 0;

        if( mga->hastuner )
                caps |= MGAVIDEO_HAS_TUNER;

        if( mga->hasaudio )
                caps |= MGAVIDEO_HAS_AUDIO;

        if( mga->hasdecoder )
                caps |= MGAVIDEO_HAS_DECODER;

        return caps;
}


u8* mgavideo_lock_video( struct mga_dev* mga )
{
        mga->locked = mga->status;

        if( mga->status & VS_FIELD ) {
                return mga->fb + _VINADDR1;
        } else {
                return mga->fb + _VINADDR0;
        }
}


void mgavideo_unlock_video( struct mga_dev* mga )
{
        mga->locked = 0;
}


/****************************************************************************
* interrupt processing
****************************************************************************/
static void interrupt_hw(int irq, void *v, struct pt_regs *regs)
{
        struct mga_dev* mga = (struct mga_dev*)v;

        if (mga == NULL)
                return;

        /* read status register */
        mga->status = _mgaread( mga, VSTATUS );
        if( !(mga->status & VI_IVSYNC) )
                return;


        /* clear interrupt pending */
        _mgawrite( mga, VICLEAR, VI_IVSYNC );


        /* increment frame counter */
        mga->grabbed++;
        

        if( !(mga->locked & ( VS_VIDEO | VS_RAWVBI | VS_SLICEDVBI ) ) ) {
                /* if no video fields are locked, program the next window
                 * based on what window was just grabbed */
                _mgawrite( mga, VINNEXTWIN, (mga->status&VS_FIELD)?0:1 );

                /* queue bottom half if present, and requested */
                if( mga->tqnode_dpc.routine ) {
                        if( ( mga->req_even && !( mga->status & VS_FIELD ) )
                        ||  ( mga->req_odd  &&  ( mga->status & VS_FIELD ) ) ) {
                                queue_task(&mga->tqnode_dpc, &tq_immediate);
                                mark_bh(IMMEDIATE_BH);
                        }
                }

        } else {
                /* otherwise, program the next window based on what field
                 * is locked */
                _mgawrite( mga, VINNEXTWIN, (mga->locked&VS_FIELD)?0:1 );

        }
}


/* enable interrupts */
void mgavideo_ivsync_enable( struct mga_dev* mga, int en )
{
        u32 istat;

        if( mga == NULL )
                return;

        dprintk( "mgavideo: %sable interrupt\n", en?"en":"dis" );

        _mgawrite( mga, VICLEAR, VI_IVSYNC );

        istat = _mgaread( mga, VIEN );
        _mgawrite( mga, VIEN, en?(istat|VI_IVSYNC):(istat&~VI_IVSYNC) );
}


/* register a routine to be called for each frame */
void mgavideo_register_bh( struct mga_dev* mga, int req_even, int req_odd,
                           void(*routine)(void*), void* data )
{
        mga->req_even = req_even;
        mga->req_odd = req_odd;
        mga->tqnode_dpc.next = NULL;
        mga->tqnode_dpc.sync = 0;
        mga->tqnode_dpc.routine = routine;
        mga->tqnode_dpc.data = data;
}


/****************************************************************************
* initialization
****************************************************************************/
static void _mgavideo_init( struct mga_dev* mga )
{
        writeb( XMISCCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) | 0x06, mga->ctrl + X_DATAREG );

        /* enable video-in */
        _mgawrite( mga, VINCTL, 0x0101ab01 );

        /* just in case, clear video interrupt */
        _mgawrite( mga, VICLEAR, VI_IVSYNC );

        /* setup video window 0 */
        _mgawrite( mga, VINCTL0, ( VIN_ON | VBI_RAW | (_VINPITCH >> 2) << 3) );
        _mgawrite( mga, VBIADDR0, _VBIADDR0 & VBIADDRx_MASK );
        _mgawrite( mga, VINADDR0, _VINADDR0 & VINADDRx_MASK );

        /* setup video window 1 */
        _mgawrite( mga, VINCTL1, ( VIN_ON | VBI_RAW | (_VINPITCH >> 2) << 3) );
        _mgawrite( mga, VBIADDR1, _VBIADDR1 & VBIADDRx_MASK );
        _mgawrite( mga, VINADDR1, _VINADDR1 & VINADDRx_MASK );

        /* setup backend scaler */
        _mgawrite( mga, BESGLOBCTL, 0x00 );
        _mgawrite( mga, BESHCOORD, (0<<16) | (0+(_VINWID-1)));
        _mgawrite( mga, BESVCOORD, (0<<16) | (0+(_VINHGT-1)));
        _mgawrite( mga, BESPITCH, _VINPITCH & (~3));
        _mgawrite( mga, BESHSRCLST, (_VINWID-1)<<16 );
        _mgawrite( mga, BESHSRCST, 0<<2 );
        _mgawrite( mga, BESHSRCEND, 703<<2 );
        _mgawrite( mga, BESV1SRCLST, 255 );
        _mgawrite( mga, BESV2SRCLST, 255 );
        _mgawrite( mga, BESA1ORG, _VINADDR0 );
        _mgawrite( mga, BESA2ORG, _VINADDR1 );
        _mgawrite( mga, BESB1ORG, _VINADDR0 );
        _mgawrite( mga, BESB2ORG, _VINADDR1 );

        /* register interrupt routine */
        if( mga->irq ) {
                if( request_irq( mga->irq, interrupt_hw, SA_SHIRQ,
                                 mga->name, mga ) < 0 ) {
                        dprintk( "mgavideo: Denied IRQ %d\n", mga->irq );
                        mga->irq_status = 0;
                } else {
                        mga->irq_status = mga->irq;
                }
        }
}


#define NBOARDS 4
struct mga_dev mgavideo_dev[NBOARDS];

/* find all mga adapters */
struct mga_dev* mgavideo_get( void )
{
        int i;
        struct mga_dev* mga;

        for( i=0; i<NBOARDS; i++ )
        {
                mga = &mgavideo_dev[i];
                if( !mga->inuse && (mga->ctrl != NULL) )
                {
                        dprintk( "mgavideo: acquire video device %x\n", i );
                        mga->inuse++;
                        MOD_INC_USE_COUNT;

                        _mgavideo_init( mga );
                        return mga;
                }
        }

        return NULL;
}

/* release a mga adapter that is not needed anymore */
void mgavideo_release( struct mga_dev* mga )
{
        if( mga == NULL )
                return;

        if( mga->ctrl )
        {
                /* clear any pending interrupt */
                _mgawrite( mga, VICLEAR, VI_IVSYNC );

                /* disable video interrupt */
                _mgawrite( mga, VIEN, _mgaread( mga, VIEN ) & ~VI_IVSYNC );

                /* disable video-in */
                _mgawrite( mga, VINCTL, 0x0101ab00 );

                /* un-register isr */
                if( mga->irq_status ) {
                        free_irq( mga->irq_status, mga );
                        mga->irq_status = 0;
                }

                MOD_DEC_USE_COUNT;
                mga->inuse--;
                dprintk( "mgavideo: release video device\n" );
                printk( "mgavideo: grabbed %d total frames\n", mga->grabbed );
        }
}


/****************************************************************************
* linux kernel module api
****************************************************************************/
#ifdef MODULE
int init_module( void )
#else
int mgavideo_init( void )
#endif
{
        int found = 0;
        struct pci_dev* dev;
        struct mga_dev* mga;
        u32 addr = 0;

        dprintk( "mgavideo: Looking for boards...\n" );

        /* find the next MGA G100 or G200 */
        for( dev = pci_devices; dev != NULL; dev = dev->next )
        {
                if (dev->class != PCI_CLASS_NOT_DEFINED_VGA &&
                  ((dev->class) >> 16 != PCI_BASE_CLASS_DISPLAY))
                        continue;

                if (PCI_FUNC(dev->devfn) != 0)
                        continue;

                /* ignore non-mga adapters */
                if (dev->vendor != PCI_VENDOR_ID_MATROX)
                        continue;

                /* ignore non-gseries adapters */
                if (dev->device != PCI_DEVICE_ID_MATROX_G200_PCI &&
                    dev->device != PCI_DEVICE_ID_MATROX_G200_AGP &&
                    dev->device != PCI_DEVICE_ID_MATROX_G100_MM &&
                    dev->device != PCI_DEVICE_ID_MATROX_G100_AGP)
                        continue;


                /* the device is cool. set up a context for it */
                mga = &mgavideo_dev[found];
                memset( mga, 0, sizeof( *mga ) );

                mga->device = dev->device;
                mga->inuse = 0;

                /* read and map the frame buffer */
                addr = dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK;
                mga->fb = ioremap( addr, 0x800000 );

                /* read and map the control aperature */
                addr = dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK;
                mga->ctrl = ioremap( addr, 0x4000 );

                /* read the device's irq */
                mga->irq = dev->irq;

                /* reset the device */
                _mgawrite( mga, VINCTL, 0x0101ab00 );

                /* setup the i2c */
                sprintf( mga->i2c.name, "mga%d", found );
                mga->i2c.id = I2C_BUSID_MGA;
                mga->i2c.data = mga;
            #if LINUX_VERSION_CODE >= 0x020100
                mga->i2c.bus_lock = SPIN_LOCK_UNLOCKED;
            #endif
                mga->i2c.attach_inform = _mgai2c_attach;
                mga->i2c.detach_inform = _mgai2c_detach;
                mga->i2c.i2c_setlines = _mgai2c_setlines;
                mga->i2c.i2c_getdataline = _mgai2c_getdataline;
                mga->i2c.i2c_read = NULL;
                mga->i2c.i2c_write = NULL;
                i2c_register_bus( &mga->i2c );

                sprintf( mga->name, "mgavideo%d", found );

                /* okay everything looks good */
                dprintk( "mgavideo: Found %s: device 0x%x irq %i\n", mga->name, mga->device, mga->irq );
                
                if( ++found >= NBOARDS )
                        break;
        }

        if( !found )
                return -ENODEV;

        return 0;
}

#ifdef MODULE
void cleanup_module( void )
{
        int i;
        struct mga_dev* mga;

        for( i=0; i<NBOARDS; i++ )
        {
                mga = &mgavideo_dev[i];
                if( mga->ctrl )
                {
                        i2c_unregister_bus( &mga->i2c );
                        iounmap( (void*)mga->ctrl );
                }
        }
}
#endif
